from pyssp_standard.transformation_types import Transformation
from pyssp_standard.common_content_ssc import Annotations, Annotation
from pyssp_standard.utils import ModelicaXMLFile
from lxml import etree as et
from lxml.etree import QName
from typing import TypedDict


class MappingEntry(TypedDict):
    source: str
    target: str
    suppress_unit_conversion: bool
    annotations: Annotations
    transformation: Transformation


class MappingList(list):

    def __repr__(self):
        print_out = \
f"""{'_'*100}
MappingList:
"""
        for item in self:
            print_out += \
f"""{'_'*20}
    Source: {item['source']}
    Target: {item['target']}
"""
        print_out += f"{'_' *100}\n"
        return print_out


class SSM(ModelicaXMLFile):

    def __init__(self, *args):
        self.version = "1.0"
        self.__mappings: MappingList[MappingEntry] = MappingList()

        super().__init__(*args, identifier='ssm')

    def __repr__(self):
        return \
f"""{'_'*100}
Parameter Mapping:
    Filepath: {self.file_path}
    Mappings: {len(self.mappings)}
{'_'*100}
"""

    def __read__(self):
        tree = et.parse(str(self.file_path))
        self.root = tree.getroot()
        self.top_level_metadata.update(self.root.attrib)
        self.base_element.update(self.root.attrib)
        self.version = self.root.get("version")

        mappings = self.root.findall('ssm:MappingEntry', self.namespaces)
        for entry in mappings:
            transformation = entry.findall('ssc:Transformation', self.namespaces)
            trans = None
            if len(transformation) > 0:
                trans_type = transformation[0].tag.split('}')[-1]
                trans = Transformation(trans_type, transformation[0].attrib)
            annotations = entry.findall('ssc:Annotations', self.namespaces)
            if len(annotations) > 0:
                annotations_list = annotations[0].findall('ssc:Annotation', self.namespaces)
                anno_list = Annotations()
                for anno in annotations_list:
                    anno_item = Annotation(type_declaration=anno.get('type'))
                    anno_item.add_element(anno)
                    anno_list.add_annotation(anno_item)

            self.__mappings.append(MappingEntry(source=entry.attrib.get('source'), target=entry.attrib.get('target'),
                                                suppress_unit_conversion=False, annotations=Annotations(),
                                                transformation=trans if trans is not None else Transformation()))

    def __write__(self):

        self.root = et.Element(QName(self.namespaces['ssm'], 'ParameterMapping'), attrib={'version': self.version})
        self.root = self.top_level_metadata.update_root(self.root)
        self.root = self.base_element.update_root(self.root)

        for mapping in self.__mappings:
            mapping_entry = et.SubElement(self.root, QName(self.namespaces['ssm'], 'MappingEntry'),
                                          attrib={'target': mapping.get('target'),
                                                  'source': mapping.get('source')})
            if mapping['transformation'] is not Transformation():
                transformation_element = mapping['transformation'].element()
                if transformation_element is not None:
                    mapping_entry.append(transformation_element)
            if not mapping['annotations'].is_empty():
                annotation_element = mapping['annotations'].root
                if annotation_element is not None:
                    mapping_entry.append(annotation_element)

    @property
    def identifier(self):
        if self.version == "2.0":
            return "ssm2"
        else:
            return "ssm"

    @property
    def mappings(self):
        return self.__mappings

    def add_mapping(self, source, target, suppress_unit_conversion=False, transformation=None, annotations=None):
        self.__mappings.append(MappingEntry(source=source, target=target,
                                            suppress_unit_conversion=suppress_unit_conversion,
                                            transformation=Transformation() if transformation is None else transformation,
                                            annotations=Annotations() if annotations is None else annotations))

    def edit_mapping(self, edit_target=True, *, target=None, source=None,
                     transformation: Transformation = None, suppress_unit_conversion=None,
                     annotations: Annotations = None):
        found = False
        idx = 0
        for idx, entry in enumerate(self.__mappings):
            if edit_target and entry.get('target') == target:
                found = True
                break
            elif not edit_target and entry.get('source') == source:
                found = True
                break

        if found:
            mapping_found = self.__mappings[idx]
            if target is not None:
                mapping_found['target'] = target
            if source is not None:
                mapping_found['source'] = source
            if transformation is not None:
                mapping_found['transformation'] = transformation
            if suppress_unit_conversion is not None:
                mapping_found['suppress_unit_conversion'] = suppress_unit_conversion
            if annotations is not None:
                mapping_found['annotations'] = annotations

        else:
            raise Exception("The target or source was not found, there is nothing to edit")



import shutil
import tempfile
import zipfile
import itertools
from datetime import datetime
from pathlib import Path, PosixPath
from dataclasses import dataclass
from lxml import etree as et

from pyssp_standard.standard import ModelicaStandard
from pyssp_standard.utils import ModelicaXMLFile, ZIPFile
from pyssp_standard.unit import Units
from pyssp_standard.common_content_ssc import (
    TypeChoice,
    Enumeration,
    TypeReal,
    TypeInteger,
    TypeBoolean,
    TypeString,
)


@dataclass
class ScalarVariable:
    name: str
    description: str
    causality: str
    variability: str
    type_: TypeChoice


class VariableList(list):

    def __repr__(self):
        print_out = \
f"""{'_'*100}
VariableList:
"""
        for item in self:
            print_out += \
f"""{'_'*20}
    Name:        {item.name}
    Description: {item.description}
    Variability: {item.variability}
    Causality:   {item.causality}
"""
        print_out += f"{'_' *100}\n"
        return print_out


class SimpleType:
    name: str
    description: str
    type_: Enumeration | TypeReal | TypeInteger | TypeBoolean | TypeString
    kind: str

    def __init__(self, name, type_, description=None, kind=None):
        self.name = name
        self.type_ = type_
        self.description = description
        self.kind = kind

    @classmethod
    def from_xml(cls, elem):
        name = elem.get("name")
        description = elem.get("description")
        type_kind = elem[0].tag
        if type_kind == "Real":
            type_ = TypeReal.from_xml(elem[0])
        elif type_kind == "Integer":
            type_ = TypeInteger.from_xml(elem[0])
        elif type_kind == "Boolean":
            type_ = TypeBoolean.from_xml(elem[0])
        elif type_kind == "String":
            type_ = TypeString.from_xml(elem[0])
        elif type_kind == "Enumeration":
            type_ = Enumeration.from_xml_fmi(name, description, elem[0])
        else:
            raise ValueError("Element is not a valid type choice element.")

        return cls(name, type_, description, type_kind)


class TypeDefinitions:
    types: list[Enumeration | TypeReal | TypeInteger | TypeBoolean | TypeString]

    def __init__(self, types):
        self.types = types

    def __len__(self):
        return len(self.types)

    def __getitem__(self, idx):
        return self.types[idx]

    def __setitem__(self, idx, val):
        self.types[idx] = val

    @classmethod
    def from_xml(cls, elem):
        children = elem.findall("SimpleType")

        return cls([SimpleType.from_xml(child) for child in children])


def _to_camel_case(snake_case):
    """Convert from snake_case to camelCase"""

    parts = snake_case.split("_")
    return "".join(itertools.chain(parts[:1], (part.capitalize() for part in parts[1:])))


class ModelDescription(ModelicaXMLFile):
    """FMI Model Description"""

    fmi_version: str
    model_name: str
    guid: str
    description: str | None
    author: str | None
    version: str | None
    copyright: str | None
    license: str | None
    generation_tool: str | None
    generation_date_and_time: datetime | None
    variable_naming_convention: str | None

    def __repr__(self):
        return \
f"""{'_'*100}
ModelDescription:
    Model Name:  {self.model_name}
    FMI Version: {self.fmi_version}
    Filepath:    {self.file_path}
{'_'*100}
"""

    def __init__(self, file_path, mode='r'):
        self.__variables: VariableList[ScalarVariable] = VariableList()
        self.type_defs = None
        self.units = None

        for name in self.__annotations__:
            setattr(self, name, None)

        super().__init__(file_path, mode, "fmi30")

    def __read__(self):
        tree = et.parse(str(self.file_path))
        root = tree.getroot()

        for name in self.__annotations__:
            xml_name = _to_camel_case(name)
            setattr(self, name, root.get(xml_name))

        if self.generation_date_and_time is not None:
            self.generation_date_and_time = datetime.fromisoformat(self.generation_date_and_time.strip("Z"))

        model_variables = root.findall('ModelVariables')[0]
        scalar_variables = model_variables.findall('ScalarVariable')
        for scalar in scalar_variables:
            name = scalar.get('name')
            description = scalar.get('description')
            causality = scalar.get('causality')
            variability = scalar.get('variability')
            type_ = TypeChoice.from_xml(scalar.xpath(TypeChoice.XPATH_FMI)[0])
            scalar_variable = ScalarVariable(
                name=name,
                description=description,
                variability=variability,
                causality=causality,
                type_=type_,
            )
            self.__variables.append(scalar_variable)

        unit_defs = root.find("UnitDefinitions")
        if unit_defs is not None:
            self.units = Units(unit_defs)

        type_defs = root.find("TypeDefinitions")
        if type_defs is not None:
            self.type_defs = TypeDefinitions.from_xml(type_defs)

    def __write__(self):
        pass

    @property
    def parameters(self) -> VariableList:
        return VariableList([item for item in self.__variables if item.causality == 'parameter'])

    @property
    def outputs(self) -> VariableList:
        return VariableList([item for item in self.__variables if item.causality == 'output'])

    @property
    def inputs(self) -> VariableList:
        return VariableList([item for item in self.__variables if item.causality == 'input'])

    def exist(self, name: str):
        """ Returns true if a scalar variable exist with the given name """
        for entry in self.__variables:
            if entry.name == name:
                return True
        return False

    def get(self, causality: str = None, variability: str = None) -> VariableList:
        """
        Get a variable from the FMU, based on the attributes' causality, variability.
        :param causality: parameter, input etc.
        :param variability: fixed, tunable etc.
        :return: list of matching variables
        """
        return VariableList([item for item in self.__variables
                             if (causality is None or item.causality == causality) and
                             (variability is None or item.variability == variability)])

    def variables(self) -> VariableList:
        return self.__variables


class FMU(ZIPFile):
    """This class allows for peeking into a FMU file, it by no means covers the entirety of the FMI standard and should
            primarily be considered as a debugging tool for looking of model input, outputs and parameters.
    """

    def __enter__(self):
        super().__enter__()
        self.fmu_binaries_path = self.unpacked_path / 'binaries'
        self.fmu_documentation_path = self.unpacked_path / 'documentation'

        return self

    def __init__(self, source_path, target_path=None, mode="a", readonly=None):
        super().__init__(source_path, target_path, mode=mode, readonly=readonly)
        self.fmu_binaries_path: Path = None
        self.fmu_documentation_path: Path = None

    def __str__(self) -> str:
        nl = "\t\n - "
        return f"""
{'_'*100}
FMU:
    Path       {self.file_path}
    Temp_dir:  {self.unpacked_path}
    Resources:{nl}{ nl.join([str(i) for i in self.files_rel])}
"""

    @property
    def model_description(self):
        md = self.unpacked_path / "modelDescription.xml"
        return ModelDescription(md)

    @property
    def binaries(self):
        """ 
        Returns a list of available binaries in the fmu folder /binaries
        """
        return [f for f in self.get_files(self.fmu_binaries_path).keys()]

    @property
    def documentation(self):
        """ 
        Returns a list of available documentation in the fmu folder /documentation
        """
        return [f for f in self.get_files(self.fmu_documentation_path).keys()]



import os
import pathlib

import hashlib
from pyssp_standard.utils import ModelicaXMLFile, XMLFile, BaseElement
from pyssp_standard.standard import ModelicaStandard
from pyssp_standard.common_content_ssc import BaseElement
from lxml import etree as et
from lxml.etree import QName


class ClassificationEntry(BaseElement, ModelicaStandard):
    """SSP Traceability ClassificationEntry

    This class represents a generic ClassificationEntry. As the standard
    is very flexible in what the "value" of a classification entry can
    be, the aim is not to fit every usecase, but rather to expose all
    data required for most common usecases. Therefore, the support for
    values other than text is only what is necessary to read and write
    in a standard-compliant manner.
    """
    keyword: str
    type_: str
    link: str | None
    linked_type: str | None
    content: list[et._Element]
    text: str

    def __init__(self, keyword_or_element: str | et._Element,
                 type_="text/plain",
                 link=None,
                 linked_type=None,
                 content: list[et._Element] | None = None,
                 text: str = "",
                 **kwargs):
        """ Construct a classification entry.

        This constructor should be called with either an lxml Element
        or keyword parameters for creating a new Classification Entry.
        Passing keyword parameters when constructing from an Element
        will result in values from kw parameters being overriden by the
        values in the XML element, or set to None if not present in the
        XML element.
        """
        super().__init__(**kwargs)

        self.content = []
        self.text = ""
        self.type_ = type_
        self.link = link
        self.linked_type = linked_type

        if isinstance(keyword_or_element, et._Element):
            self.content = []
            self.__read__(keyword_or_element)
        elif isinstance(keyword_or_element, str):
            self.keyword = keyword_or_element
            if content is not None:
                self.content = content
            self.text = text
        else:
            raise TypeError(f"Can't init ClassificationEntry with {type(keyword_or_element)}")

    def __read__(self, element):
        self.update({"id": element.get("id", ""), "description": element.get("description", "")})

        self.keyword = element.attrib["keyword"]
        self.type_ = element.get("type", "text/plain")
        self.link = element.get(QName(self.namespaces["xlink"], "href"))
        self.linked_type = element.get("linkedType")

        if element.text is not None:
            self.text = element.text

        for child in element:
            self.content.append(child)

    def __repr__(self):
        return f"ClassificationEntry(keyword={self.keyword}, text={self.text!r} content={self.content})"

    def as_element(self):
        entry = et.Element(QName(self.namespaces['stc'], 'ClassificationEntry'), attrib={'keyword': self.keyword})
        entry.text = self.text
        entry.extend(self.content)

        super().update_root(entry)

        if self.type_ != "text/plain":  # Only write if the value differs from the default
            entry.set("type", self.type_)

        if self.link is not None:  # optional
            entry.set(QName(self.namespaces["xlink"], "href"), self.link)

        if self.linked_type is not None:  # optional
            entry.set("linkedType", self.linked_type)

        return entry


class Classification(BaseElement, ModelicaStandard):
    """SSP Traceability Classification.

    This class represents a generic Classification. For easiser parsing
    of specific Classification types, this class can be subclassed, and
    registered with the @classification_parser decorator. A possible
    way to implement this is shown in the documentation for that
    decorator.
    """
    classification_type: str
    link: str
    linked_type: str
    classification_entries: list[ClassificationEntry]

    def __init__(self,
                 type_or_element: str | et._Element,
                 link: str | None = None,
                 linked_type: str | None = None,
                 entries: list[ClassificationEntry] | None = None,
                 **kwargs):
        """ Construct a classification.

        This constructor should be called with either an lxml Element
        or keyword parameters for creating a new Classification.
        Passing keyword parameters when constructing from an Element
        will result in values from kw parameters being overriden by the
        values in the XML element, or set to None if not present in the
        XML element.
        """
        super().__init__(**kwargs)

        self.link = link
        self.linked_type = linked_type
        self.classification_entries = []

        if isinstance(type_or_element, et._Element):
            self.__read__(type_or_element)
        elif isinstance(type_or_element, str):
            self.classification_type = type_or_element
            self.classification_entries = [] if entries is None else entries
        else:
            raise TypeError(f"Can't init Classification with {type(type_or_element)}")

    def add_classification_entry(self, classification_entry: ClassificationEntry):
        self.classification_entries.append(classification_entry)

    def __read__(self, element):
        # BaseElement doesn't work well with inheritance
        self.update({"id": element.get("id", ""), "description": element.get("description", "")})

        self.classification_type = element.attrib["type"]
        self.link = element.get(QName(self.namespaces["xlink"], "href"))
        self.linked_type = element.get("linkedType")

        for entry in element.getchildren():
            self.add_classification_entry(ClassificationEntry(entry))

    def as_element(self):
        classification = et.Element(QName(self.namespaces['stc'], 'Classification'),
                                    attrib={'type': self.classification_type})

        super().update_root(classification)

        if self.link is not None:
            classification.set(QName(self.namespaces["xlink"], "href"), self.link)

        if self.linked_type is not None:
            classification.set("linkedType", self.link)

        for entry in self.classification_entries:
            classification.append(entry.as_element())

        return classification


classification_parsers: dict[str, Classification] = {}


def classification_parser(type_: str):
    """Decorator for registering a classification parser for a given type.

    The parser class' constructor will be called with the XML element as
    its only argument. The class is expected to have a method as_element(self)
    that returns the XML element representation of the parsed classification.

    For example: to parse the following classification schema
    <stc:Classification type="com.example.custom">
        <stc:ClassicationEntry keyword="test1">Value 1</stc:ClassificationEntry>
        <stc:ClassicationEntry keyword="test2">Value 2</stc:ClassificationEntry>
    </stc:Classification>
    one might do something like the following code snippet.

    >>> @classification_parser("com.example.custom")
    >>> class CustomClassification(Classification):
    >>>     test1: str
    >>>     test2: str

    >>>     def __init__(self, element=None, test1="", test2="", **kwargs):
    >>>         if element is not None:
    >>>             super().__init__(element)

    >>>             for entry in self.classification_entries:
    >>>                 if entry.keyword == "test1":
    >>>                     self.test1 = entry.text
    >>>                 elif entry.keyword == "test2":
    >>>                     self.test2 = entry.text
    >>>         else:
    >>>             super().__init__("com.example.custom", **kwargs)
    >>>             self.test1 = test1
    >>>             self.test2 = test2

    >>>     def as_element(self):
    >>>         self.classification_entries = [
    >>>                 ClassificationEntry("test1", text=self.test1),
    >>>                 ClassificationEntry("test2", text=self.test2)
    >>>         ]

    >>>         return super().as_element()
    """
    def decorator(parser):
        register_classification_parser(type_, parser)

        return parser

    return decorator


def classification_parser_for(type_: str) -> type:
    return classification_parsers.get(type_, Classification)


def register_classification_parser(type_: str, parser: type):
    classification_parsers[type_] = parser


class SRMD(ModelicaXMLFile):

    def __init__(self, file_path, mode='r'):
        self.name = os.path.basename(file_path)
        self.classifications = []
        self.data = None
        self.checksum = None
        self.checksum_type = "SHA3-256"
        self.version = "1.0.0-beta2"

        super().__init__(file_path, mode, "srmd11")

    def assign_data(self, filepath, create_checksum=True):
        if type(filepath) is not pathlib.PosixPath:
            filepath = pathlib.Path(filepath)
        self.data = str(filepath)

        if create_checksum:
            with open(filepath) as file:
                data = file.read()
                self.checksum = hashlib.sha3_256(data.encode()).hexdigest()

    def __read__(self):
        tree = et.parse(str(self.file_path))
        self.root = tree.getroot()
        self.version = self.root.get('version')
        self.name = self.root.get('name')
        self.data = self.root.get('data')
        self.checksum = self.root.get('checksum')
        self.checksum_type = self.root.get('checksumType')

        self.top_level_metadata.update(self.root.attrib)
        self.base_element.update(self.root.attrib)

        classifications = self.root.findall('stc:Classification', self.namespaces)
        for classification in classifications:
            type_ = classification.attrib['type']
            cls = classification_parser_for(type_)  # select an appropriate parser based on the registry
            self.add_classification(cls(classification))

    def __write__(self):
        attributes = {'version': self.version, 'name': self.name}
        if self.data is not None:
            attributes['data'] = self.data
        if self.checksum is not None:
            attributes['checksum'] = self.checksum
            attributes['checksumType'] = self.checksum_type

        self.root = et.Element(
                QName(self.namespaces['srmd'], 'SimulationResourceMetaData'),
                attrib=attributes,
        )

        self.root = self.top_level_metadata.update_root(self.root)
        self.root = self.base_element.update_root(self.root)
        for classification in self.classifications:
            self.root.append(classification.as_element())

        et.cleanup_namespaces(self.root, top_nsmap=self.namespaces)

    def add_classification(self, classification: Classification):
        self.classifications.append(classification)



from pyssp_standard.ssd import SSD, Connection
from pyssp_standard.ssb import SSB
from pyssp_standard.ssv import SSV
from pyssp_standard.ssm import SSM
from pyssp_standard.fmu import FMU
from pyssp_standard.ssp import SSP
from pyssp_standard.transformation_types import Transformation
from pyssp_standard.common_content_ssc import Annotations, Annotation

from pyssp_standard.srmd import SRMD, Classification, ClassificationEntry, classification_parser



import shutil
import filecmp
import tempfile
from pathlib import Path, PosixPath
from abc import ABC, abstractmethod
import zipfile
import xmlschema
import os
import warnings
from lxml import etree as ET

from pyssp_standard.common_content_ssc import Annotations, Annotation, BaseElement, TopLevelMetaData
from pyssp_standard.standard import ModelicaStandard


def register_namespaces():
    for name, url in ModelicaStandard.namespaces.items():
        ET.register_namespace(name, url)


class XMLFile(ABC):
    """
    Base for all xml files
    """

    @abstractmethod
    def __read__(self):
        """
        Fill class with information from xml file
        """
        pass

    @abstractmethod
    def __write__(self):
        """
        Write class data to xml objec
        """
        pass

    @property
    def BaseElement(self):
        return self.base_element

    @property
    def TopLevelMetaData(self):
        return self.top_level_metadata

    def __init__(self, file_path, mode="r"):
        """
        :param mode: [r]ead, [a]ppend or [w]rite
        """
        self.__mode = mode
        if type(file_path) is not PosixPath:
            file_path = Path(file_path)

        self.__file_path = file_path
        self.root = None
        self.base_element: BaseElement = BaseElement()
        self.top_level_metadata: TopLevelMetaData = TopLevelMetaData()

        if mode == "r" or mode == "a":
            self.__read__()

    def write_to_file(self, filepath):
        xml_string = ET.tostring(
            self.root, pretty_print=True, encoding="utf-8", xml_declaration=True
        )
        with open(filepath, "wb") as file:
            file.write(xml_string)

    def check_compliance(self, schema, namespaces):
        # check name for indications that xsd 1.1 should be used, 1.0 is the default
        if "11" in str(schema):
            schema = xmlschema.XMLSchema11(schema)
        else:
            schema = xmlschema.XMLSchema10(schema)

        if self.__mode in ["a", "w"]:  # Temporary file creation
            with tempfile.TemporaryDirectory(suffix="_pyssp") as temp_dir:
                temp_file_path = Path(temp_dir) / "tmp.xml"
                self.__write__()
                self.write_to_file(temp_file_path)

                xmlschema.validate(temp_file_path, schema, namespaces=namespaces)
        else:
            xmlschema.validate(self.file_path, schema, namespaces=namespaces)

    @property
    def file_path(self):
        return self.__file_path

    def __save__(self):
        """
        Write xml object to file
        """
        self.write_to_file(self.__file_path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__mode in ["w", "a"]:
            self.__write__()
            self.__save__()


class ModelicaXMLFile(XMLFile, ModelicaStandard):
    """
    Base Class for all spp xml files
    """

    def __read__(self):
        pass

    def __write__(self):
        pass

    def __check_compliance__(self):
        super().check_compliance(self.schemas[self.identifier], self.namespaces)

    @property
    def identifier(self):
        return self.__identifier

    @property
    def annotations(self):
        return self.__annotations

    def add_annotation(self, annotation: Annotation):
        self.annotations.add_annotation(annotation)

    def __enter__(self):
        register_namespaces()
        return self

    def __init__(self, file_path, mode="r", identifier="unknown"):
        self.__identifier = identifier
        self.__annotations = Annotations()
        super().__init__(file_path, mode)


def _is_file_content_equal(path: Path | str, content: str | bytes) -> bool:
    """Return true if file specified by path is equal to content."""
    mode = "r" if isinstance(content, str) else "rb"
    with open(path, mode) as f:
        return f.read() == content


class ZIPFile:
    """
    All operations need to be in context and will be applied against temp dir

    """

    def __enter__(self):
        self.__temp_path = Path(tempfile.mkdtemp(prefix="pyssp_"))
        self.__unpacked_path = self.__temp_path / self.file_path.stem

        if self.mode == "r" or (self.mode == "a" and Path(self.file_path).exists()):
            with zipfile.ZipFile(self.file_path, "r") as zip_ref:
                zip_ref.extractall(self.__unpacked_path)
        else:
            self.__unpacked_path.mkdir()

        self.__in_context = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.mode != "r" and self.__changed:
            zip_file_path = shutil.make_archive(self.__unpacked_path, "zip", self.__unpacked_path)
            shutil.copy(zip_file_path, self.save_path)

        shutil.rmtree(self.__temp_path)

        self.__in_context = False

    def __init__(self, source_path: Path, target_path: Path = None, mode="a", readonly=None):
        """
        If target_path is not specified it will overwrite the opened file at exit.
        This can be probibited by specifying mode="r".

        Opening modes:
            * [a]ppend: open and read the contents of an SSP archive,
              creating an empty archive if the file doesn't exist
            * [w]rite: create an empty SSP archive to write to, overriding the target_path
            * [r]ead: read the contents of an SSP archive without modifying
              the contents.

        The readonly parameter is deprecated (but remains for backwards-compatibility).
        Migrating can be done by changing readonly=True to mode='r' and readonly=False
        to mode='a'.
        """
        if type(source_path) is not PosixPath:
            source_path = Path(source_path)

        self.file_path = source_path
        self.save_path = target_path if target_path is not None else source_path

        # Remove once the readonly parameter is removed!
        if readonly is not None:
            if mode != "a":
                raise ValueError("The mode parameter makes readonly redundant. Specify only mode!")

            mode = "r" if readonly else "a"
            type_name = type(self).__name__
            message = (
                f"The readonly parameter is deprecated. Use {type_name}(..., {mode=})"
                f"instead of {type_name}(..., {readonly=}) to maintain current behavior"
            )
            warnings.warn(message, DeprecationWarning, stacklevel=3)

        self.mode = mode

        self.__changed = False
        self.__in_context = False
        self.__temp_path = ""
        self.__unpacked_path = ""

    def mark_changed(self):
        self.__changed = True

    def check_context(self):
        if not self.__in_context:
            raise Exception("Function or variable not accessable unless opened using 'with'")

    @staticmethod
    def get_files(dir: Path):
        return {os.path.relpath(p, dir): p for p in dir.rglob("*")}

    @property
    def unpacked_path(self):
        self.check_context()
        return self.__unpacked_path

    @property
    def __files(self):
        """
        get at dict with [rel_path:abs_path]
        """
        return self.get_files(self.__unpacked_path)

    @property
    def files_rel(self):
        """
        get a list of files relative to zip file
        """
        self.check_context()
        return self.__files.keys()

    @property
    def files_abs(self):
        """
        get a list of files, with absolute path within the temp dir
        """
        self.check_context()
        return self.__files.values()

    def get_file_temp_path(self, rel_path):
        """
        translate from rel_path to abs_path

        """
        self.check_context()
        return self.__unpacked_path / rel_path

    def add_file(self, file: Path, rel_path="", overwrite=False):
        """
        Add something to the resource folder of the ssp.
        :param file: filepath of the object to add.
        :param rel_path: relative path inside archive, omitt '/' in the beginning
        :param overwrite: if True, overwrite file in archive if it already exists
        """
        self.check_context()
        if self.mode == "r":
            raise Exception("Changes are not allowed in readonly archive")

        self.mark_changed()
        # Create subdirectory if it doesn't already exist
        archive_dir = self.get_file_temp_path(rel_path)
        archive_dir.mkdir(parents=True, exist_ok=True)  # eqv. to mkdir -p ...

        rel_path = Path(rel_path) / Path(file).name

        temp_path = self.get_file_temp_path(rel_path)
        if overwrite or not temp_path.exists() or filecmp.cmp(file, temp_path, shallow=False):
            shutil.copy(file, temp_path)
        else:
            # This shouldn't fail silently
            raise FileExistsError(f"File {rel_path} already exists in archive")

    def add_file_contents(self, content: str | bytes, rel_path: Path, overwrite=False):
        """
        Add something to the resource folder of the ssp.
        :param content: Contents of the file to write as a str or bytes object.
        :param rel_path: relative path inside archive, omitt '/' in the beginning
        :param overwrite: if True, overwrite file in archive if it already exists
        """
        self.check_context()
        if self.mode == "r":
            raise Exception("Changes are not allowed in readonly archive")

        self.mark_changed()
        # Create subdirectory if it doesn't alreeshutil.copy(file, temp_path)dy exist
        archive_dir = self.get_file_temp_path(rel_path.parent)
        archive_dir.mkdir(parents=True, exist_ok=True)  # eqv. to mkdir -p ...

        temp_path = self.get_file_temp_path(rel_path)
        if overwrite or not temp_path.exists() or _is_file_content_equal(temp_path, content):
            with open(temp_path, "w" if isinstance(content, str) else "wb") as f:
                f.write(content)
        else:
            # This shouldn't fail silently
            raise FileExistsError(f"File {rel_path} already exists in archive")

    def remove_file(self, rel_path):
        """
        relative path should be within the file, no slash in the begining
        e.g. SystemStructure.ssd
        """
        self.check_context()
        if self.mode == "r":
            raise Exception("Changes are not allowed in readonly archive")

        self.mark_changed()
        file: Path = self.__unpacked_path / rel_path

        if file.exists():
            file.unlink()
        else:
            raise FileNotFoundError(f"Not found {file}")


class SSPElement(ABC):
    @abstractmethod
    def to_element(self) -> ET.Element:
        pass

    @abstractmethod
    def from_element(self, element: ET.Element):
        pass


class EmptyElement(ET.ElementBase):
    def __init__(self, tag, attrib=None, **kwargs):
        super().__init__(tag, attrib, **kwargs)

    def __repr__(self):
        return ET.tostring(self, encoding="utf-8")



from dataclasses import dataclass, asdict

from lxml import etree as ET
from lxml.etree import QName

from pyssp_standard.utils import SSPElement
from pyssp_standard.standard import ModelicaStandard


@dataclass
class BaseUnit:
    kg: int = None
    m: int = None
    s: int = None
    A: int = None
    K: int = None
    mol: int = None
    cd: int = None
    rad: int = None
    factor: float = None
    offset: float = None

    def __init__(self, base_unit: dict):
        for field_name, field_type in self.__annotations__.items():
            value = base_unit.get(field_name)
            if value is not None:
                if not isinstance(value, field_type):
                    try:
                        value = field_type(value)
                    except (TypeError, ValueError):
                        raise ValueError(
                            f"Invalid value type for {field_name}. Expected {field_type}."
                        )
                setattr(self, field_name, value)

    def to_dict(self):
        return {k: str(v) for k, v in asdict(self).items() if v is not None}


class Unit(SSPElement, ModelicaStandard):
    def __init__(self, unit, base_unit: BaseUnit = None):
        self.name = None
        self.base_unit = None

        if isinstance(unit, ET._Element):
            self.from_element(unit)
        else:
            self.name = unit
            self.base_unit = base_unit

    def to_element(self, namespace="ssc"):
        ns = "" if not namespace else f"{{{self.namespaces[namespace]}}}"

        unit_entry = ET.Element(f"{ns}Unit", attrib={"name": self.name})

        if namespace == "ssc" and self.base_unit is None:
            self.base_unit = BaseUnit({})  # In FMI BaseUnit is optional, in SSP it is required

        if self.base_unit is not None:
            ET.SubElement(unit_entry, f"{ns}BaseUnit", **self.base_unit.to_dict())

        return unit_entry

    def from_element(self, element):
        self.name = element.attrib.get("name")
        ns = QName(element.tag).namespace
        tag_name = "ssc:BaseUnit" if ns is not None else "BaseUnit"

        base_unit = element.find(tag_name, self.namespaces)
        if base_unit is not None:  # Base unit is optional
            self.base_unit = BaseUnit(base_unit.attrib)


class Units(ModelicaStandard):
    def __init__(self, element: ET.Element = None):
        self.units = {}

        if element is not None:
            namespace = "ssc:"
            if element.tag == "UnitDefinitions":  # Detect FMI ModelDefinition
                namespace = ""

            units = element.findall(f"{namespace}Unit", self.namespaces)
            for unit_elem in units:
                unit = Unit(unit_elem, namespace)
                self.units[unit.name] = unit

    def add_unit(self, unit: Unit):
        """Add unit to units listing.

        If unit already exists but with a different definition a ValueError is raised.
        Otherwise, update listing to include the passed unit, overriding base unit if
        not already in listing.
        """
        if unit.name in self:
            own = self.units[unit.name]

            # Same name, different base unit
            if (
                own.base_unit is not None
                and unit.base_unit is not None
                and own.base_unit != unit.base_unit
            ):
                raise ValueError(f"Unit {unit.name} already exists with different definition")

            own.base_unit = own.base_unit or unit.base_unit
        else:
            self.units[unit.name] = unit

    def element(self, parent_type="ssc"):
        if parent_type == "fmi":
            elem_name = "UnitDefinitions"
            child_ns = None
        else:
            elem_name = QName(self.namespaces[parent_type], "Units")
            child_ns = "ssc"

        root = ET.Element(elem_name)
        root.extend(unit.to_element(namespace=child_ns) for unit in self.units.values())

        return root

    def __len__(self):
        return len(self.units)

    def __iter__(self):
        return iter(self.units.values())

    def __getitem__(self, key):
        return self.units[key]

    def __contains__(self, key):
        return key in self.units

    def is_empty(self):
        return len(self.units) == 0




from pyssp_standard.parameter_types import ParameterType
from pyssp_standard.common_content_ssc import Annotations, Enumerations, Enumeration
from pyssp_standard.unit import Units
from pyssp_standard.utils import ModelicaXMLFile
from lxml import etree as ET
from lxml.etree import QName
from typing import TypedDict


class DictionaryEntry(TypedDict):
    name: str
    type_entry: ParameterType
    annotations: Annotations


class DictionaryEntryList(list):

    def __repr__(self):
        print_out = \
f"""{'_' *100}
DictionaryEntryList:"
"""
        for item in self:
            print_out += \
f"""{'_' *20}
    Name: {item['name']}
    Type: {item['type_entry']}
"""
        print_out += f"{'_' *100}\n"
        return print_out


class SSB(ModelicaXMLFile):

    def __init__(self, *args):
        self.version = None

        self.__dictionary_entry: DictionaryEntryList = DictionaryEntryList()
        self.__enumerations: Enumerations = Enumerations()
        self.__units: Units = Units()

        super().__init__(*args, identifier='ssb')

    def __read__(self):
        tree = ET.parse(str(self.file_path), parser=ET.XMLParser(encoding='utf-8'))
        self.root = tree.getroot()

        self.version = self.root.get('version')
        dictionary_entries = self.root.findall('ssb:DictionaryEntry', self.namespaces)
        for entry in dictionary_entries:
            name = entry.get('name')
            self.__dictionary_entry.append(DictionaryEntry(name=name, type_entry=ParameterType(),
                                                           annotations=Annotations()))

    def __write__(self):
        self.root = ET.Element(QName(self.namespaces['ssb'], 'SignalDictionary'), attrib={'version': '1.0'})
        self.root = self.top_level_metadata.update_root(self.root)
        self.root = self.base_element.update_root(self.root)
        for entry in self.__dictionary_entry:
            dictionary_entry = ET.Element(QName(self.namespaces['ssb'], 'DictionaryEntry'), attrib={'name': entry.get('name')})
            dictionary_entry.append(entry["type_entry"].element())
            self.root.append(dictionary_entry)

    @property
    def identifier(self):
        if self.version == "2.0":
            return "ssb2"
        else:
            return "ssb"

    def add_dictionary_entry(self, name: str, ptype: str, value: dict, annotations=None):
        self.__dictionary_entry.append(DictionaryEntry(name=name,
                                                       type_entry=ParameterType(ptype, value, 'ssc'),
                                                       annotations=Annotations() if annotations is None else annotations))

    def add_enumeration(self, enum: Enumeration):
        self.__enumerations.add_enumeration(enum)

    def add_unit(self):
        pass



from collections import defaultdict
from dataclasses import dataclass, field

from pyssp_standard.common_content_ssc import (
    Enumerations,
    Annotations,
    Annotation,
    TypeChoice,
    TypeReal,
    BaseElement
)
from pyssp_standard.unit import Units
from pyssp_standard.utils import ModelicaXMLFile
from pyssp_standard.standard import ModelicaStandard
from pyssp_standard.ssv import SSVElem
from lxml import etree as ET
from lxml.etree import QName

_ALLOWED_CONNECTIONS = {
    ("System", "parameter", "System", "calculatedParameter"),
    ("System", "parameter", "System", "output"),
    ("System", "input", "System", "output"),
    ("System", "parameter", "Element", "parameter"),
    ("System", "parameter", "Element", "input"),
    ("System", "parameter", "Element", "inout"),
    ("System", "input", "Element", "input"),
    ("System", "input", "Element", "inout"),
    ("Element", "calculatedParameter", "Element", "parameter"),
    ("Element", "calculatedParameter", "Element", "input"),
    ("Element", "calculatedParameter", "Element", "inout"),
    ("Element", "output", "Element", "input"),
    ("Element", "output", "Element", "inout"),
    ("Element", "inout", "Element", "input"),
    ("Element", "calculatedParameter", "System", "calculatedParameter"),
    ("Element", "calcualtedParameter", "System", "output"),
    ("Element", "output", "System", "output"),
    ("Element", "inout", "System", "output"),
}

_DESTINATIONS = {
    ("Element", "input"),
    ("Element", "parameter"),
    ("System", "output"),
    ("System", "calculatedParameter"),
}


class Connection(ModelicaStandard):

    def __init__(self, element=None, *, start_element=None, start_connector=None, end_element=None, end_connector=None):
        self.__root = None

        self.base_element = None
        self.start_element = start_element  # Optional
        self.start_connector = start_connector
        self.end_element = end_element  # Optional
        self.end_connector = end_connector
        self.suppress_unit_conversion: bool = False

        self.transformation = None
        self.annotations = None
        # Connection geometry not connected

        if element is not None:
            self.__read__(element)

    def __read__(self, element):
        self.start_element = element.get('startElement')
        self.start_connector = element.get('startConnector')
        self.end_element = element.get('endElement')
        self.end_connector = element.get('endConnector')

    def __eq__(self, other):
        if self.start_element == other.start_element and self.start_connector == other.start_connector and \
                self.end_element == other.end_element and self.end_connector == other.end_connector:
            return True
        else:
            return False

    def as_element(self) -> ET.Element:
        attrib = {
                'startElement': self.start_element,
                'startConnector': self.start_connector,
                'endElement': self.end_element,
                'endConnector': self.end_connector
        }
        attrib = {k: v for k, v in attrib.items() if v is not None}

        self.__root = ET.Element(QName(self.namespaces['ssd'], 'Connection'), **attrib)
        return self.__root

    def as_dict(self):
        return {'source': self.start_element, 'signal': self.start_connector,
                'target': self.end_element, 'receiver': self.end_connector}

    def __repr__(self) -> str:
        return f"""source {self.start_element} - {self.start_connector} -> target {self.end_element} - {self.end_connector}"""


class Connector(ModelicaStandard):

    def __init__(self, element=None, name="", kind="", type_=TypeReal(None)):
        self.name = name
        self.kind = kind
        self.type_ = type_

        if element is not None:
            self.__read__(element)

    def __read__(self, element: ET.Element):
        self.name = element.get('name')
        self.kind = element.get('kind')
        type_elem = TypeChoice.XPATH_SSP(element)
        self.type_ = TypeChoice.from_xml(type_elem[0]) if type_elem else None

    def as_element(self):
        elem = ET.Element(
                QName(self.namespaces["ssd"], "Connector"),
                name=self.name,
                kind=self.kind
        )
        elem.append(self.type_.to_xml(namespace="ssc"))

        return elem

    def as_dict(self):
        return {'name': self.name, 'kind': self.kind}


class Component(ModelicaStandard):
    def __init__(self, element=None):
        self.component_type = None
        self.name = None
        self.source = None
        self.implementation = None
        self.connectors = []
        self.parameter_bindings = None
        self.annotations = None

        if element is not None:
            self.__read__(element)

    def __read__(self, element):
        self.name = element.get('name')
        self.component_type = element.get('type')
        self.source = element.get('source')
        self.implementation = element.get('implementation')

        connectors = element.find('ssd:Connectors', namespaces=self.namespaces)
        if connectors is not None:
            for connector in connectors.findall('ssd:Connector', namespaces=self.namespaces):
                self.connectors.append(Connector(connector))

        self.parameter_bindings = element.find("ssd:ParameterBindings", namespaces=self.namespaces)

    def as_element(self):
        element = ET.Element(QName(self.namespaces["ssd"], "Component"), name=self.name)

        if self.component_type is not None:
            element.set("type", self.component_type)

        if self.source is not None:
            element.set("source", self.source)

        if self.implementation is not None:
            element.set("implementation", self.implementation)

        if self.connectors:
            connectors = ET.Element(QName(self.namespaces["ssd"], "Connectors"))
            for connector in self.connectors:
                connectors.append(connector.as_element())

            element.append(connectors)

        if self.parameter_bindings is not None:
            element.append(self.parameter_bindings)

        return element

    def as_dict(self):
        return {'name': self.name, 'connectors': [connector.as_dict() for connector in self.connectors]}


class Element(ModelicaStandard):

    def __init__(self, element):
        self.components = []
        self.__read__(element)

    def __read__(self, element):
        components = element.findall('ssd:Component', namespaces=self.namespaces)
        for component in components:
            self.components.append(Component(component))

    def as_dict(self):
        return [component.as_dict() for component in self.components]


@dataclass
class ParameterBinding(ModelicaStandard):
    base_element: BaseElement = field(default_factory=BaseElement)
    type: str = "application/x-ssp-parameter-set"
    source: str | None = None
    source_base: str = "SSD"
    prefix: str | None = None
    ssv: SSVElem | None = field(default_factory=SSVElem)
    # TODO: support for parameter mapping

    @classmethod
    def from_xml(cls, elem):
        base_element = BaseElement()
        base_element.update(elem.attrib)

        type_ = elem.get("type", "application/x-ssp-parameter-set")
        source = elem.get("source")
        source_base = elem.get("sourceBase", "SSD")
        prefix = elem.get("prefix")
        parameter_values = elem.find("ssd:ParameterValues", cls.namespaces)
        ssv = None
        if parameter_values is not None:
            ssv = parameter_values.find("ssv:ParameterSet", cls.namespaces)

        return cls(base_element, type_, source, source_base, prefix, ssv)

    def to_xml(self):
        root = ET.Element(
            QName(self.namespaces['ssd'], 'ParameterBinding'),
        )
        self.base_element.update(root.attrib)
        if self.type != "application/x-ssp-parameter-set":
            root.set("type", self.type)

        if self.source is not None:
            root.set("source", self.source)

        if self.source_base != "SSD":
            root.set("sourceBase", self.source_base)

        if self.prefix is not None:
            root.set("prefix", self.prefix)

        if self.ssv is not None:
            parameter_values = ET.SubElement(root, QName(self.namespaces["ssd"], "ParameterValues"))
            parameter_values.append(self.ssv.to_xml())

        return root


class System(ModelicaStandard):
    name: str
    elements: "list[Component | System]"  # ugly, because of forward declarations. Fixed in py3.14+
    connections: list[Connection]

    connectors: list[Connector]
    parameter_bindings: list[ParameterBinding]
    signal_dictionaries: list
    annotations: Annotations | None

    def __init__(
            self,
            system_element: ET.Element = None,
            name: str = "",
        ):
        self.name = name
        self.elements = []
        self.connections: list[Connection] = []

        self.connectors: list[Connector] = []
        self.parameter_bindings = []
        self.signal_dictionaries = []
        self.annotations = Annotations(namespace="ssd")

        if system_element is not None:
            self.__read__(system_element)

    def parse_element(self, elem):
        if elem.tag == QName(self.namespaces["ssd"], "Component"):
            return Component(elem)
        elif elem.tag == QName(self.namespaces["ssd"], "System"):
            return System(elem)
        else:
            # Unfortunately, no support for SignalDictionaries yet :(
            return elem

    def __read__(self, element):
        self.name = element.get('name', None)

        connectors = element.find('ssd:Connectors', namespaces=self.namespaces)
        if connectors is not None:
            for connector in connectors.findall('ssd:Connector', namespaces=self.namespaces):
                self.connectors.append(Connector(connector))

        elements = element.find('ssd:Elements', namespaces=self.namespaces)
        if elements is not None:
            self.elements = [self.parse_element(child) for child in elements]

        connections = element.find('ssd:Connections', namespaces=self.namespaces)
        if connections is not None:
            for connection in connections.findall('ssd:Connection', namespaces=self.namespaces):
                self.connections.append(Connection(connection))

        annotations = element.find("ssd:Annotations", namespaces=self.namespaces)
        if annotations is not None:
            for elem in annotations:
                self.annotations.add_annotation(Annotation(elem))

        parameter_bindings = element.find("ssd:ParameterBindings", namespaces=self.namespaces)
        if parameter_bindings is not None:
            bindings = parameter_bindings.findall("ssd:ParameterBinding", namespaces=self.namespaces)
            self.parameter_bindings = [ParameterBinding.from_xml(binding) for binding in bindings]

    def as_element(self):
        element = ET.Element(QName(self.namespaces["ssd"], "System"), name=self.name)

        if self.connectors:
            connectors = ET.Element(QName(self.namespaces["ssd"], "Connectors"))
            connectors.extend(connector.as_element() for connector in self.connectors)
            element.append(connectors)

        if self.parameter_bindings:
            parameter_bindings = ET.SubElement(
                element,
                QName(self.namespaces["ssd"], "ParameterBindings")
            )
            parameter_bindings.extend(binding.to_xml() for binding in self.parameter_bindings)

        if self.elements:
            elements = ET.Element(QName(self.namespaces["ssd"], "Elements"))
            elements.extend(el.as_element()
                  if isinstance(el, (Component, System))
                  else el for el in self.elements)
            element.append(elements)

        if self.connections:
            connections = ET.Element(QName(self.namespaces["ssd"], "Connections"))
            connections.extend(connection.as_element() for connection in self.connections)
            element.append(connections)

        if not self.annotations.is_empty():
            element.append(self.annotations.element())

        return element

    def check_connections(
        self,
        unallowed_connections=True,
        ambiguous_data_flow=True,
        unconnected_inputs=False,
        connector_not_in_system=False
    ):
        """Check connection semantics

        Args:
        unallowed_connections: if True, return warnings about
            connections involving combinations not allowed by the SSP
            standard, e.g. element output -> element output.
        ambiguous_data_flow: if True, return warnings about input
            connectors having multiple inbound connections.
        unconnected_inputs, if True, return warnings about input
            connectors lacking any inbound connections. False by default
            since this is not an error according to the SSP standard.
        connector_not_in_system: if True, return warnings about connection
            start or end connector not being present in the system.


        Return: list of warning strings
        """
        connector_kind: dict[tuple[str, str], str] = {}
        connector_owner: dict[tuple[str, str], str] = {}

        inbound = {}

        for connector in self.connectors:
            connector_kind[(None, connector.name)] = connector.kind
            connector_owner[(None, connector.name)] = "System"

            if ("System", connector.kind) in _DESTINATIONS:
                inbound[("", connector.name)] = []

        for element in self.elements:
            for connector in element.connectors:
                connector_kind[(element.name, connector.name)] = connector.kind
                connector_owner[(element.name, connector.name)] = "Element"

                if ("Element", connector.kind) in _DESTINATIONS:
                    inbound[(element.name, connector.name)] = []

        warnings = []

        for connection in self.connections:
            source = (connection.start_element, connection.start_connector)
            dest = (connection.end_element, connection.end_connector)
            source_owner_kind = (connector_owner[source], connector_kind[source])
            dest_owner_kind = (connector_owner[dest], connector_kind[dest])

            if (*source_owner_kind, *dest_owner_kind) in _ALLOWED_CONNECTIONS:
                pass  # Allowed connection, data flow from start -> end
            elif (*dest_owner_kind, *source_owner_kind) in _ALLOWED_CONNECTIONS:
                # Allowed connection, data flow from end -> start
                source, dest = dest, source
                source_owner_kind, dest_owner_kind = dest_owner_kind, source_owner_kind
            elif unallowed_connections:
                warnings.append(f"Unallowed connection combination: {source[0]}.{source[1]} "
                    f"({source_owner_kind[0]} {source_owner_kind[1]}) -> "
                    f"{dest[0]}.{dest[1]} ({dest_owner_kind[0]} {dest_owner_kind[1]})")

            if source not in connector_kind and connector_not_in_system:
                warnings.append(f"Source connector not found in system for connection: "
                    f"{source[0]}.{source[1]} -> {dest[0]}.{dest[1]}")
            if dest not in connector_kind and connector_not_in_system:
                warnings.append(f"Destination connector not found in system for connection: "
                    f"{source[0]}.{source[1]} -> {dest[0]}.{dest[1]}")

            if dest in inbound:
                inbound[dest].append(source)

        for connector, inbound_connectors in inbound.items():
            if len(inbound_connectors) == 0 and unconnected_inputs:
                warnings.append("Input connector has no inbound connections: "
                    f"{connector[0]}.{connector[1]}")
            elif len(inbound_connectors) > 1 and ambiguous_data_flow:
                connectors = ", ".join(f"{con[0]}.{con[1]}" for con in inbound_connectors)
                warnings.append(f"Input connector {connector[0]}.{connector[1]} has ambiguous "
                f"dataflow (multiple inbound connections): {connectors}")

        return warnings


class DefaultExperiment(ModelicaStandard):

    def __init__(self, element: ET.Element = None):
        self.start_tim: float = None
        self.stop_time: float = None
        self.annotations: Annotations = Annotations(namespace="ssd")

        if element is not None:
            self.__read__(element)

    def __read__(self, element):
        self.start_time = float(element.get('startTime'))
        self.stop_time = float(element.get('stopTime'))

        annotations = element.find('ssd:Annotations', self.namespaces)

        if annotations is not None:
            for annotation in annotations.findall('ssc:Annotation', self.namespaces):
                self.annotations.add_annotation(Annotation(annotation))

    def as_element(self):
        elem = ET.Element(
                QName(self.namespaces["ssd"], "DefaultExperiment"),
                startTime=str(self.start_time),
                stopTime=str(self.stop_time)
        )

        if not self.annotations.is_empty():
            elem.append(self.annotations.element())

        return elem


class SSD(ModelicaXMLFile):

    def __init__(self, file_path, mode='r'):

        self.name = None
        self.version = None

        self.system = None
        self.default_experiment = None
        self.enumerations: Enumerations = Enumerations(namespace="ssd")
        self.units: Units = Units()

        super().__init__(file_path=file_path, mode=mode, identifier='ssd')

    def __read__(self):
        parser = ET.XMLParser(remove_blank_text=True)
        tree = ET.parse(str(self.file_path), parser)
        self.root = tree.getroot()

        self.top_level_metadata.update(self.root.attrib)
        self.base_element.update(self.root.attrib)

        system = self.root.find('ssd:System', self.namespaces)
        if system is not None:
            self.system = System(system)

        default_experiment = self.root.findall('ssd:DefaultExperiment', self.namespaces)
        if len(default_experiment) > 0:
            self.default_experiment = DefaultExperiment(default_experiment[0])

        enumerations = self.root.find("ssd:Enumerations", self.namespaces)
        if enumerations is not None:
            self.enumerations = Enumerations(enumerations, namespace="ssd")

        units = self.root.find("ssd:Units", self.namespaces)
        if units is not None:
            self.units = Units(units)

        self.name = self.root.get('name')
        self.version = self.root.get('version')

    def __write__(self):
        namespaces = ["ssd", "ssc"]
        nsmap = {k: self.namespaces[k] for k in namespaces}
        self.root = ET.Element(
                QName(self.namespaces["ssd"], "SystemStructureDescription"),
                version=self.version,
                name=self.name,
                nsmap=nsmap
        )

        self.root = self.top_level_metadata.update_root(self.root)
        self.root = self.base_element.update_root(self.root)

        if self.system is not None:
            self.root.append(self.system.as_element())

        if self.default_experiment is not None:
            self.root.append(self.default_experiment.as_element())

        if self.enumerations is not None and self.enumerations.enumerations:
            self.root.append(self.enumerations.as_element())

        if self.units is not None and len(self.units) != 0:
            self.root.append(self.units.element(parent_type="ssd"))

    @property
    def identifier(self):
        if self.version == "2.0":
            return "ssd2_11"
        else:
            return "ssd11"

    def check_connections(self, **kwargs):
        return self.system.check_connections(**kwargs)

    def add_connection(self, connection: Connection):
        if type(connection) is not Connection:
            raise "Only Connection object may be used."
        self.system.connections.append(connection)

    def remove_connection(self, connection: Connection):
        try:
            self.system.connections.remove(connection)
        except ValueError:
            pass  # Replicate previous behavior

    def connections(self):
        return self.system.connections

    def list_connections(self, *, start_connector=None, end_connector=None, start_element=None, end_element=None):
        connections = self.connections()
        matching_connections = []

        def check(value, comparision):
            return True if value is None else value == comparision

        for connection in connections:
            connection : Connection
            sc = [start_connector, connection.start_connector]
            ec = [end_connector, connection.end_connector]
            se = [start_element, connection.start_element]
            ee = [end_element, connection.end_element]

            if check(*sc) and check(*ec) and check(*se) and check(*ee):
                matching_connections.append(connection)

        return matching_connections

    def list_connectors(self, *, kind=None, name=None, parent=None):
        """Returns a list of connectors, filtered by the following optional options
            :param kind: the kind of connector, e.g. input, output or parameter
            :param name: the name of the connector, utilizes 'in' for lookup
            :param parent: the name of the parent component, utilizes 'in' for lookup
        """

        matching_connectors = defaultdict(list)

        for element in self.system.elements:
            if not isinstance(element, (System, Component)):
                continue

            if parent is not None and parent not in element.name:
                continue

            for connector in element.connectors:
                if kind is not None and kind != connector.kind:
                    continue
                if name is not None and name not in connector.name:
                    continue

                matching_connectors[element.name].append(
                        {"name": connector.name, "kind": connector.kind})

        return matching_connectors



from typing import TypedDict
from lxml import etree as ET
from lxml.etree import QName

from pyssp_standard.standard import ModelicaStandard


class Transformation(ModelicaStandard):

    def __init__(self, transformation_type=None, attributes=None, transformation: ET.Element = None):

        self.transformation_type = None
        self.transformation = None
        if transformation is not None:
            self.transformation_type = transformation.tag.split('}')[-1]
            self.transformation = self.__create_transformation__(self.transformation_type, transformation.attrib)
        elif transformation_type is not None and attributes is not None:
            self.transformation_type = transformation_type
            self.transformation = self.__create_transformation__(transformation_type, attributes)

    def element(self):
        if self.transformation_type is not None:
            converted_transformation = {key: str(value) for key, value in self.transformation.items()}
            return ET.Element(QName(self.namespaces['ssc'], self.transformation_type), attrib=converted_transformation)
        return None

    @staticmethod
    def __create_transformation__(ttype, attributes):
        source = attributes.get('source')
        target = attributes.get('target')
        factor = attributes.get('factor')
        offset = attributes.get('offset')

        transformation_types = {
            'LinearTransformation': lambda: LinearTransformation(factor=factor, offset=offset),
            'BooleanMappingTransformation': lambda: BooleanMappingTransformation(target=target, source=source),
            'EnumerationMappingTransformation': lambda: EnumerationMappingTransformation(target=target, source=source),
            'IntegerMappingTransformation': lambda: IntegerMappingTransformation(target=target, source=source)
        }

        return transformation_types.get(ttype, lambda: None)()


class TransformationType(TypedDict):
    pass


class LinearTransformation(TransformationType):
    factor: float  # default 1.0
    offset: float  # default 0.0


class BooleanMappingTransformation(TransformationType):
    source: bool
    target: bool


class IntegerMappingTransformation(TransformationType):
    source: int
    target: int


class EnumerationMappingTransformation(TransformationType):
    source: str
    target: str





import pint


def generate_base_unit(input_string):

    ureg = pint.UnitRegistry()
    parsed_quantity = ureg(input_string)

    conversion_table = {'meter': 'm',
                        'kilogram': 'kg',
                        'second': 's',
                        'ampere': 'A',
                        'kelvin': 'K',
                        'mole': 'mol',
                        'candela': 'Cd',
                        'radian': 'rad'
                        }
    unit_dict = {}
    units_to_parse = (str(parsed_quantity.to_base_units())).split()
    unit_dict['factor'] = float(units_to_parse[0])
    if len(units_to_parse) == 2 and units_to_parse[1] == 'kelvin':
        unit_dict = {'offset': float(units_to_parse[0]), conversion_table[units_to_parse[1]]: 1}
    elif len(units_to_parse) == 2:
        unit_dict[conversion_table[units_to_parse[1]]] = 1
    else:
        idx = 1
        previous_operator = None
        while idx < len(units_to_parse):
            if units_to_parse[idx] in ['*', '/']:
                previous_operator = units_to_parse[idx]
                change = 1
            else:
                unit = units_to_parse[idx]
                change = 2 if (idx + 1 < len(units_to_parse) - 1 and units_to_parse[idx + 1] in ['**', '*', '/']) else 1

                if '**' in units_to_parse[idx:idx + 3]:
                    exponent = int(units_to_parse[idx + 2])
                    change = 3
                else:
                    exponent = 1

                exponent_factor = -1 if previous_operator == '/' else 1
                unit_dict[conversion_table[unit]] = exponent_factor * exponent
                previous_operator = units_to_parse[idx + 1] if idx + 1 < len(units_to_parse) else None
            idx += change

    return unit_dict



import os
import tempfile
import zipfile
import shutil
from pathlib import Path, PosixPath
import warnings
from pyssp_standard.ssd import SSD
from pyssp_standard.ssb import SSB
from pyssp_standard.ssv import SSV
from pyssp_standard.ssm import SSM
from pyssp_standard.fmu import FMU
from pyssp_standard.standard import ModelicaStandard
from pyssp_standard.utils import ZIPFile


class VariantsProxy:
    archive_root: Path
    mode: str

    def __init__(self, archive_root: Path, mode: str):
        self.archive_root = archive_root
        self.mode = mode

    def __len__(self):
        return len(self.archive_root.glob("*.ssd"))

    def __contains__(self, name):
        variant_path = self.archive_root / Path(name).with_suffix(".ssd")
        return variant_path.is_file()

    def __getitem__(self, name):
        if name not in self and self.mode == "r":
            raise KeyError(f"SSD archive has no variant named {name!r}")

        variant_path = self.archive_root / Path(name).with_suffix(".ssd")

        mode = self.mode
        if mode == "a":
            mode = "a" if variant_path.exists() else "w"

        return SSD(variant_path, mode=self.mode)

    def __iter__(self):
        return (path.stem for path in self.archive_root.glob("*.ssd"))


class SSP(ZIPFile):
    def __enter__(self):
        super().__enter__()
        self.ssp_resource_path = self.unpacked_path / "resources"

        return self

    def __init__(self, source_path, target_path=None, mode="a", readonly=None):
        super().__init__(source_path, target_path, mode=mode, readonly=readonly)
        self.ssp_resource_path: Path = None

    def __rep__(self) -> str:
        spacing = "\t\t"
        return f"""{"_" * 100}
SSP:
    Path       {self.file_path}
    Temp_dir:  {self.unpacked_path}
    Resources:
{spacing}{spacing.join([str(r) for r in self.resources])}
{"_" * 100}
"""

    @property
    def system_structure(self):
        self.mark_changed()
        ssd_path = self.get_file_temp_path("SystemStructure.ssd")

        if self.mode == "r" and not ssd_path.exists():
            raise FileNotFoundError("SystemStructure.ssd not found in unpacked archive")

        mode = self.mode
        if mode == "a":
            mode = "a" if ssd_path.exists() else "w"

        return SSD(ssd_path, mode=mode)

    @property
    def variants(self):
        self.mark_changed()
        return VariantsProxy(self.unpacked_path, self.mode)

    @property
    def ssd(self):
        message = (
            "The ssd property is deprecated. Use SSP.system_structure to access the default "
            "variant, and SSP.variants to access other variant SSDs."
        )
        warnings.warn(message, DeprecationWarning)

        self.mark_changed()
        ssd = list(self.unpacked_path.glob("*.ssd"))[0]
        return SSD(ssd)

    @property
    def ssv(self):
        self.mark_changed()
        ssv = list(self.ssp_resource_path.glob("*.ssv"))
        return [SSV(ssv) for ssv in ssv]

    @property
    def ssm(self):
        self.mark_changed()
        ssm = list(self.ssp_resource_path.glob("*.ssm"))
        return [SSM(file) for file in ssm]

    @property
    def ssb(self):
        self.mark_changed()
        ssb = list(self.ssp_resource_path.glob("*.ssb"))
        return [SSB(file) for file in ssb]

    @property
    def fmu(self):
        self.mark_changed()
        fmu = list(self.ssp_resource_path.glob("*.fmu"))
        return [FMU(file) for file in fmu]

    @property
    def resources(self):
        """
        Returns a list of available resources in the ssp folder /resources
        """
        return [f for f in self.get_files(self.ssp_resource_path).keys()]

    def add_resource(self, file: Path, overwrite: bool | None = None):
        """
        Add something to the resource folder of the ssp.
        :param file: filepath of the object to add.
        :param overwrite: if True, overwrite resource, if False,
            FileExistsError is raised if attempting to add resource
            which already exists in the archive.
        """
        kwargs = {"overwrite": overwrite} if overwrite is not None else {}
        self.add_file(file, "resources", **kwargs)

    def remove_resource(self, resource_name):
        """
        Remove resource
        """
        if type(resource_name) is not str:
            resource_name = resource_name.name

        self.remove_file(f"resources/{resource_name}")



from typing import TypedDict
from lxml import etree as ET
from lxml.etree import QName

from pyssp_standard.standard import ModelicaStandard


class ParameterType(ModelicaStandard):

    def __init__(self, parameter_type=None, attributes=None, namespace='ssv'):
        self.namespace = namespace

        if type(parameter_type) is ET.Element:
            self.parameter_type = parameter_type.tag.split('}')[-1]
            self.parameter = self.__create_parameter__(self.parameter_type, parameter_type.attrib)
        elif type(parameter_type) is str:
            self.parameter_type = parameter_type
            self.parameter = self.__create_parameter__(parameter_type, attributes)
        else:
            raise Exception(f"No parameter type is provided attribs: {attributes}")

    def element(self):
        if self.namespace == 'ssc':
            return ET.Element(QName(self.namespaces[self.namespace], self.parameter_type))
        else:
            return ET.Element(QName(self.namespaces[self.namespace], self.parameter_type), attrib=self.parameter)

    @staticmethod
    def __create_parameter__(ptype, attributes):
        value = attributes.get('value')
        name = attributes.get('name')
        unit = attributes.get('unit')
        mimetype = attributes.get('mem-type')

        parameter_types = {
            'Real': lambda: Real(value=value, unit=unit),
            'Integer': lambda: Integer(value=value),
            'Boolean': lambda: Boolean(value=value),
            'String': lambda: String(value=value),
            'Enumeration': lambda: Enumeration(value=value, name=name),
            'Binary': lambda: Binary(value=value, mimetype=mimetype)
        }

        return parameter_types.get(ptype, lambda: None)()


class ParameterValueType(TypedDict):
    pass


class Real(ParameterValueType):
    value: float
    unit: str


class Integer(ParameterValueType):
    value: int


class Boolean(ParameterValueType):
    value: bool


class String(ParameterValueType):
    value: str


class Enumeration(ParameterValueType):
    value: str
    name: str


class Binary(ParameterValueType):
    mimetype: str  # default 'application/octet-stream'
    value: str  # should be hexadecimal



import datetime
from typing import TypedDict, List
from abc import ABC, abstractmethod
from lxml import etree as ET

from lxml.etree import QName
from dataclasses import dataclass, asdict, fields, field

from pyssp_standard.standard import ModelicaStandard


class Annotation(ModelicaStandard):  # TODO needs to read and not just create
    def __init__(self, type_declaration):
        """
        The SSP standard allows for the addition of annotations, when created they must contain at least one annotation.
        An annotation may contain anything, however to ease its use the pyssp_standard provides tools to add text, attributes
        elements and ET.Element.
        :param type_declaration: normalized string
        """
        if type(type_declaration) is str:
            self.root = ET.Element(
                QName(self.namespaces["ssc"], "Annotation"), attrib={"type": type_declaration}
            )
        elif type(type_declaration) is ET._Element:
            self.root = type_declaration

    def add_element(self, element: ET.Element):
        self.root.append(element)

    def add_text(self, text: str):
        self.root.text = text

    def add_dict(self, name: str, attributes: dict):
        self.root.append(ET.Element(name, attrib=attributes))


class Annotations(ModelicaStandard):
    def __init__(self, namespace="ssc"):
        self.__count = 0
        self.root = ET.Element(QName(self.namespaces[namespace], "Annotations"))

    def add_annotation(self, annotation: Annotation):
        self.__count += 1
        self.root.append(annotation.root)

    def element(self):
        return self.root

    def is_empty(self):
        return True if self.__count == 0 else False


@dataclass
class BaseElement:
    # Note: default should ideally be switched to None
    # <Element description="" /> is different from <Element />
    id: str = ""
    description: str = ""

    def __repr__(self):
        return f"ID: {self.id} \nDescription: {self.description}"

    def update(self, attributes: dict):
        for key, value in attributes.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def dict(self):
        return {key: str(value) for key, value in asdict(self).items()}

    def update_root(self, root: ET.Element):
        for key, value in self.dict().items():
            if value != "":
                root.set(key, value)
        return root


@dataclass
class TopLevelMetaData:
    author: str = ""
    fileversion: str = ""
    copyright: str = ""
    license: str = ""
    generationTool: str = "pyssp_standard"
    generationDateAndTime: datetime.datetime = datetime.datetime.now().isoformat()

    def __repr__(self):
        base_txt = ""
        for field in fields(self):
            reformatted_text = field.name.replace("_", " ")
            reformatted_text = reformatted_text.capitalize()
            base_txt += f"{reformatted_text}: {getattr(self, field.name)}\n"
        return base_txt

    def update(self, attributes: dict):
        for key, value in attributes.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def dict(self):
        return {key: str(value) for key, value in asdict(self).items()}

    def update_root(self, root):
        for key, value in self.dict().items():
            if value != "":
                root.set(key, value)
        return root


@dataclass
class Item(ModelicaStandard):
    name: str
    value: int

    @classmethod
    def from_xml(cls, elem):
        return cls(elem.get("name"), int(elem.get("value")))

    def to_xml(self):
        return ET.Element(
            QName(self.namespaces["ssc"], "Item"), name=self.name, value=str(self.value)
        )


@dataclass
class Enumeration(ModelicaStandard):
    name: str
    items: List[Item]
    base_element: BaseElement = field(default_factory=BaseElement)
    annotations: Annotations | None = None

    @classmethod
    def from_xml(cls, elem):
        base_elem = BaseElement()
        base_elem.update(elem.attrib)

        name = elem.get("name")
        items = [Item.from_xml(el) for el in elem.findall("ssc:Item", cls.namespaces)]
        # TODO: Annotations

        return cls(name, items, base_elem, None)

    @classmethod
    def from_xml_fmi(cls, name, description, elem):
        base_elem = BaseElement()
        if description is not None:
            base_elem.description = description

        items = [Item.from_xml(el) for el in elem.findall("Item")]
        # TODO: Annotations

        return cls(name, items, base_elem, None)

    def to_xml(self):
        elem = ET.Element(QName(self.namespaces["ssc"], "Enumeration"), name=self.name)
        self.base_element.update_root(elem)
        elem.extend(item.to_xml() for item in self.items)

        return elem


class Enumerations(ModelicaStandard):
    enumerations: list[Enumeration]
    namespace: str

    def __init__(self, enumerations: List[Enumeration] | ET.ElementBase = None, namespace="ssc"):
        self.enumerations = []
        self.namespace = namespace
        if isinstance(enumerations, ET._Element):
            self.__read__(enumerations)
        elif enumerations is not None:
            self.enumerations.extend(enumerations)

    def __read__(self, element):
        self.enumerations.extend(Enumeration.from_xml(elem) for elem in element)

    def as_element(self):
        elem = ET.Element(QName(self.namespaces[self.namespace], "Enumerations"))
        for enum in self.enumerations:
            elem.append(enum.to_xml())

        return elem

    def add_enumeration(self, enum: Enumeration):
        self.enumerations.append(enum)

    def is_empty(self):
        return len(self.enumerations) == 0


_prefix = f"{{{ModelicaStandard.namespaces['ssc']}}}"


class TypeChoice(ABC, ModelicaStandard):
    XPATH_SSP = ET.ETXPath(
        "|".join(
            _prefix + type_ for type_ in ["Real", "Integer", "Boolean", "String", "Enumeration"]
        )
    )

    XPATH_FMI = "Real|Integer|Boolean|String|Enumeration"

    @abstractmethod
    def to_xml(self, namespace="ssc"): ...

    @classmethod
    def from_xml(cls, elem):
        name = QName(elem.tag).localname
        if name == "Real":
            return TypeReal.from_xml(elem)
        elif name == "Integer":
            return TypeInteger.from_xml(elem)
        elif name == "Boolean":
            return TypeBoolean.from_xml(elem)
        elif name == "String":
            return TypeString.from_xml(elem)
        elif name == "Enumeration":
            return TypeEnumeration.from_xml(elem)
        else:
            raise ValueError("Element is not a valid type choice element.")


class TypeReal(TypeChoice):
    def __init__(self, unit, min=None, max=None, start=None):
        self.unit = unit
        self.min = min
        self.max = max
        self.start = start

    def to_xml(self, namespace="ssc"):
        if namespace:
            ns = f"{{{self.namespaces[namespace]}}}"
        else:
            ns = ""

        elem = ET.Element(f"{ns}Real")
        if self.unit is not None:
            elem.set("unit", self.unit)

        if namespace == "":  # fmi
            if self.min is not None:
                elem.set("min", str(self.min))

            if self.max is not None:
                elem.set("max", str(self.max))

            if self.start is not None:
                elem.set("start", str(self.start))

        return elem

    @classmethod
    def from_xml(cls, elem):
        keys = ("min", "max", "start")

        return cls(
            unit=elem.get("unit"),
            **{key: float(v) for key in keys if (v := elem.get(key)) is not None},
        )


class TypeInteger(TypeChoice):
    def __init__(self):
        pass

    def to_xml(self, namespace="ssc"):
        if namespace:
            ns = f"{{{self.namespaces[namespace]}}}"
        else:
            ns = ""

        return ET.Element(f"{ns}Integer")

    @classmethod
    def from_xml(cls, elem):
        return cls()


class TypeBoolean(TypeChoice):
    def __init__(self):
        pass

    def to_xml(self, namespace="ssc"):
        if namespace:
            ns = f"{{{self.namespaces[namespace]}}}"
        else:
            ns = ""

        return ET.Element(f"{ns}Boolean")

    @classmethod
    def from_xml(cls, elem):
        return cls()


class TypeString(TypeChoice):
    def __init__(self):
        pass

    def to_xml(self, namespace="ssc"):
        if namespace:
            ns = f"{{{self.namespaces[namespace]}}}"
        else:
            ns = ""

        return ET.Element(f"{ns}String")

    @classmethod
    def from_xml(cls, elem):
        return cls()


class TypeEnumeration(TypeChoice):
    def __init__(self, name):
        self.name = name

    def to_xml(self, namespace="ssc"):
        kwargs = {}
        if namespace:
            ns = f"{{{self.namespaces[namespace]}}}"
            kwargs["name"] = self.name
        else:  # fmi
            ns = ""
            kwargs["declaredType"] = self.name

        return ET.Element(f"{ns}Enumeration", **kwargs)

    @classmethod
    def from_xml(cls, elem):
        # Parse fmi enums with attr declaredType
        ns = QName(elem.tag).namespace
        if ns is None:  # fmi
            name = elem.get("declaredType")
        else:  # ssp
            name = elem.get("name")
        return cls(name)



from pathlib import Path


class ModelicaStandard:
    namespaces = {
        # SSP
        "ssc": "http://ssp-standard.org/SSP1/SystemStructureCommon",
        "ssv": "http://ssp-standard.org/SSP1/SystemStructureParameterValues",
        "ssb": "http://ssp-standard.org/SSP1/SystemStructureSignalDictionary",
        "ssm": "http://ssp-standard.org/SSP1/SystemStructureParameterMapping",
        "ssd": "http://ssp-standard.org/SSP1/SystemStructureDescription",
        # SRMD
        "stc": "http://ssp-standard.org/SSPTraceability1/SSPTraceabilityCommon",
        "srmd": "http://ssp-standard.org/SSPTraceability1/SimulationResourceMetaData",
        # FMI
        "fmi30": "",
        # XLink
        "xlink": "http://www.w3.org/1999/xlink"
    }

    __resource_path = Path(__file__).parent / "resources"
    schemas = {
        # SSP
        "ssc": __resource_path / "SystemStructureCommon.xsd",
        "ssd": __resource_path / "SystemStructureDescription.xsd",
        "ssd11": __resource_path / "SystemStructureDescription11.xsd",
        "ssm": __resource_path / "SystemStructureParameterMapping.xsd",
        "ssv": __resource_path / "SystemStructureParameterValues.xsd",
        "ssb": __resource_path / "SystemStructureSignalDictionary.xsd",
        # SSP 2.0
        "ssc2": __resource_path / "ssp2" / "SystemStructureCommon.xsd",
        "ssd2": __resource_path / "ssp2" / "SystemStructureDescription.xsd",
        "ssd2_11": __resource_path / "ssp2" / "SystemStructureDescription11.xsd",
        "ssm2": __resource_path / "ssp2" / "SystemStructureParameterMapping.xsd",
        "ssv2": __resource_path / "ssp2" / "SystemStructureParameterValues.xsd",
        "ssb2": __resource_path / "ssp2" / "SystemStructureSignalDictionary.xsd",
        # SSPTraceabillity
        "stc11": __resource_path / "STC11.xsd",
        "srmd11": __resource_path / "SRMD11.xsd",
        # FMI
        "fmi30": __resource_path / "fmi30" / "fmi3ModelDescription.xsd",
    }



from distutils.core import setup

setup(
    name='pyssp_standard',
    packages=['pyssp_standard'],
    version='0.9.0',
    license='MIT',
    description='Simple python package for reading, modifying and creating files, specified in the SSP Standard',
    long_description='',
    long_description_content_type='text/markdown',
    author='Fredrik Haider',
    author_email='',
    url='https://github.com/pyssporg/pyssp_standard',
    download_url='https://github.com/pyssporg/pyssp_standard/releases/latest/',
    keywords=['SSP', 'system', 'engineering'],
    install_requires=[
        'lxml',
        'xmlschema',
        'pint',
    ],
    package_data={"pyssp_standard": ["resources/*.xsd", "resources/fmi30/*.xsd", "resources/ssp2/*.xsd"]},
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)



from dataclasses import dataclass, field, fields
from typing import TypedDict, List
from operator import attrgetter

from lxml.etree import QName
from lxml import etree as ET

from pyssp_standard.common_content_ssc import Enumerations, BaseElement, TopLevelMetaData
from pyssp_standard.parameter_types import ParameterType

from pyssp_standard.unit import BaseUnit, Unit, Units
from pyssp_standard.utils import ModelicaXMLFile
from pyssp_standard.standard import ModelicaStandard
from pyssp_standard.unit_conversion import generate_base_unit


class Parameter(TypedDict):
    name: str
    type_name: str
    type_value: ParameterType


@dataclass
class SSVElem(ModelicaStandard):
    name: str = "default"
    version: str = "1.0"
    parameters: list[Parameter] = field(default_factory=list)
    units: Units | None = field(default_factory=Units)
    enumerations: Enumerations | None = field(default_factory=Enumerations)
    base_element: BaseElement = field(default_factory=BaseElement)
    top_level_metadata: TopLevelMetaData = field(default_factory=TopLevelMetaData)

    @classmethod
    def from_xml(cls, elem):
        parameters = []

        parameters_elem = elem.find('ssv:Parameters', cls.namespaces)
        if parameters_elem is not None:
            parameter_elems = parameters_elem.findall('ssv:Parameter', cls.namespaces)
            for parameter in parameter_elems:
                name = parameter.get('name')
                param = parameter[0]
                param_type = QName(param.tag).localname
                param_attr = ParameterType(param_type, param.attrib)
                parameters.append(Parameter(name=name, type_name=param_type, type_value=param_attr))

        version = elem.get("version")
        name = elem.get("name")
        base_element = BaseElement()
        top_level_metadata = TopLevelMetaData()

        base_element.update(elem.attrib)
        top_level_metadata.update(elem.attrib)

        units_elem = elem.find('ssv:Units', cls.namespaces)
        units = Units(units_elem) if units_elem is not None else None

        enums_elem = elem.find("ssv:Enumerations", cls.namespaces)
        enumerations = Enumerations(enums_elem, namespace="ssv") if enums_elem is not None else None

        return cls(name, version, parameters, units, enumerations)

    def to_xml(self):
        namespaces = ["ssv", "ssc"]
        nsmap = {k: self.namespaces[k] for k in namespaces}
        root = ET.Element(
            QName(self.namespaces['ssv'], 'ParameterSet'),
            nsmap=nsmap,
            attrib={'version': self.version, 'name': self.name}
        )
        self.base_element.update(root.attrib)
        self.top_level_metadata.update(root.attrib)

        parameters_entry = ET.SubElement(root, QName(self.namespaces['ssv'], 'Parameters'))
        for param in self.parameters:
            parameter_entry = ET.SubElement(parameters_entry, QName(self.namespaces['ssv'], 'Parameter'),
                                            attrib={'name': param.get('name')})
            parameter_entry.append(param["type_value"].element())

        if not self.units.is_empty():
            root.append(self.units.element('ssv'))

        if self.enumerations is not None and not self.enumerations.is_empty():
            root.append(self.enumerations.as_element())

        return root


def wraps_dataclass(wrapped_class, local_name):
    def decorator(cls):
        for field_ in fields(wrapped_class):
            name = field_.name

            def setter(self, value):
                setattr(self, f"{local_name}.{name}", value)

            setattr(cls, name, property(attrgetter(f"{local_name}.{name}"), setter))

        return cls

    return decorator


@wraps_dataclass(SSVElem, local_name="ssv_elem")
class SSV(ModelicaXMLFile):
    name: str
    version: str
    parameters: list[Parameter]
    units: Units | None
    enumerations: Enumerations | None

    def __read__(self):
        tree = ET.parse(str(self.file_path))
        self.root = tree.getroot()
        self.ssv_elem = SSVElem.from_xml(self.root)

    def __write__(self):
        self.root = self.ssv_elem.to_xml()

    def __init__(self, filepath, mode='r', name='unnamed'):
        self.ssv_elem = SSVElem()
        super().__init__(file_path=filepath, mode=mode, identifier='ssv')

    @property
    def identifier(self):
        if self.version == "2.0":
            return "ssv2"
        else:
            return "ssv"

    def add_parameter(self, parname: str, ptype: str = 'Real', *, value: float = None, name: str = None, mimetype=None,
                      unit: str = None):
        parameter_dict = {}
        if value is not None:
            parameter_dict['value'] = str(value)
        if name is not None:
            parameter_dict['name'] = name
        if mimetype is not None:
            parameter_dict['mimetype'] = mimetype
        if unit is not None:
            parameter_dict['unit'] = unit
        p = Parameter(name=parname, type_name=ptype, type_value=ParameterType(ptype, parameter_dict))
        self.parameters.append(p)

    def add_unit(self, name: str, base_unit: dict = None):
        """
        Add a unit definition to the .ssv file. If base_unit is None, an attempt is made to automatically generate a BaseUnit.
        :param name: e.g. N or J/kg/K
        :param base_unit: A dictionary declaring the base unit as specified by the SSP standard.
        """
        if base_unit is None:
            base_unit = generate_base_unit(name)
        self.units.add_unit(Unit(name, base_unit=BaseUnit(base_unit)))
